!========================================================================
!                       LAMBDA OPERATOR MODULE
!
! For the Lambda Iteration scheme (and its sister, the Accelerated
! Lambda Iteration scheme) a Lambda Operator is required.
!
! A Lambda operator is simply an operator (subroutine) that computes
! the mean intensity at some wavelength using the integrals over a large
! number of rays. In contrast to the mean intensity subroutine in the
! montecarlo_module, for the Lambda operator we do not "follow a photon
! along its way" but we instead "integrate the formal transfer equation
! along a ray". If no an-isotropic dust scattering is included in the
! problem, and the scattering source function is known, this gives a 
! more accurate answer. Also, depending on which method of Lambda
! operator you use, you get less noise. Finally, one can use the
! Accelerated Lambda Iteration scheme to speed up convergence of the
! Lamber Iteration.
!========================================================================
module lambda_module
use mathroutines_module
use rtglobal_module
use amr_module
use amrray_module
use constants_module

double precision :: lambda_shift = 1.d90
integer :: lambda_nfreq,lambda_istar
double precision, allocatable :: lambda_frequencies(:)
double precision, allocatable :: lambda_intensity(:),lambda_bgint(:)
logical :: lambda_secondorder

double precision, allocatable :: lambda_vertex_snu(:)
double precision, allocatable :: lambda_vertex_anu(:)
double precision, allocatable :: lambda_cell_snu(:)
double precision, allocatable :: lambda_cell_anu(:)

contains


!------------------------------------------------------------------------
!                         INIT THIS MODULE 
!------------------------------------------------------------------------
subroutine lambda_init(nf,freq,secondorder,bgint)
  implicit none
  integer :: nf
  double precision :: freq(1:nf)
  double precision, optional :: bgint(1:nf)
  logical :: secondorder
  lambda_nfreq = nf
  allocate(lambda_intensity(1:nf),lambda_bgint(1:nf),lambda_frequencies(1:nf))
  lambda_intensity(:) = 0.d0
  lambda_frequencies(:) = freq(:)
  if(present(bgint)) then
     lambda_bgint(:) = bgint(:)
  else
     lambda_bgint(:) = 0.d0
  endif
  allocate(lambda_cell_snu(1:nrcellsmax))
  allocate(lambda_cell_anu(1:nrcellsmax))
  if(secondorder) then
     allocate(lambda_vertex_snu(1:amr_nr_vertices_max))
     allocate(lambda_vertex_anu(1:amr_nr_vertices_max))
  endif
end subroutine lambda_init


!------------------------------------------------------------------------
!                        CLOSE THIS MODULE
!------------------------------------------------------------------------
subroutine lambda_cleanup()
  implicit none
  if(allocated(lambda_intensity)) deallocate(lambda_intensity)
  if(allocated(lambda_bgint)) deallocate(lambda_bgint)
  if(allocated(lambda_frequencies)) deallocate(lambda_frequencies)
  if(allocated(lambda_cell_snu)) deallocate(lambda_cell_snu)
  if(allocated(lambda_cell_anu)) deallocate(lambda_cell_anu)
  if(allocated(lambda_vertex_snu)) deallocate(lambda_vertex_snu)
  if(allocated(lambda_vertex_anu)) deallocate(lambda_vertex_anu)
end subroutine lambda_cleanup



!------------------------------------------------------------------------
!              HELPER ROUTINE FOR RT INTEGRATION STEP
!------------------------------------------------------------------------
####################
####################
subroutine lambda_integrate_element()
  implicit none
  integer :: inu
  !
  ! Check if we are/were in a star
  !
  if(lambda_istar.le.0) then
     !
     ! Normal situation: we are not inside a star (istar=0), or we will
     ! hit the star but are not inside the star yet (istar<0).
     !
     ! Decide whether we do 1st order or 2nd order integration
     !
     if(lambda_secondorder) then
        !
        ! Second order integration. Corner-based (=vertex-based) scheme.
        ! Here all sources have all been pre-calculated on
        ! the cell corners. 
        !
        inu = inu0
           !
           ! Backup snu and anu at previous crossing
           !
           snu_prev = snu_curr
           anu_prev = anu_curr
           jnu_prev = jnu_curr
           !
           ! Find snu and anu at new crossing
           !
           ! NOTE: If lambda_interpol_jnu is set, then snu is actually
           !       j_nu (the emissivity). It will be translated below.
           !
           if(igrid_type.lt.100) then 
              if(igrid_coord.lt.100) then
                 !
                 ! Cartesian
                 !
                 call lambda_find_srcalp_interpol(       &
                         ray_cart_x,ray_cart_y,ray_cart_z,  &
                         snu_curr,anu_curr)
                 elseif(igrid_coord.lt.200) then
                    !
                    ! Spherical 
                    !
                    call amr_xyz_to_rthphi(ray_cart_x,ray_cart_y,ray_cart_z, &
                                           r,theta,phi)
                    call lambda_find_srcalp_interpol(    &
                         r,theta,phi,snu_curr,anu_curr)
                 else
                    stop 498
                 endif
              else
                 stop 499
              endif
              !
              ! If lambda_interpol_jnu is set, then what we get back
              ! from the camera_find_srcalp_interpol() is in fact the
              ! emissivity j_nu instead of the source function S_nu.
              ! So translate either snu to jnu or vice versa.
              !
              if(lambda_interpol_jnu) then
                 jnu_curr = snu_curr
                 snu_curr = jnu_curr / (anu_curr+1d-99)
              else
                 jnu_curr = snu_curr * (anu_curr+1d-99)
              endif
              !
              if(ray_index.gt.0) then
                 dtau1     = 0.5d0 * ( anu_prev + anu_curr ) * ray_ds
                 theomax   = 0.5d0 * ( jnu_prev + jnu_curr ) * ray_ds
                 if(dtau1.gt.1.d-6) then
                    xp = exp(-dtau1)
                    e0 = 1.d0 - xp
                    e1 = dtau1 - e0
                    bx = e1 / dtau1
                    ax = e0 - bx
                 else
                    ax = 0.5d0 * dtau1
                    bx = 0.5d0 * dtau1
                    xp = 1.d0 - dtau1 
                 endif
                 if(dtau1.gt.1e-9) then
                    qdr = ax * snu_prev + bx * snu_curr
                 else
                    qdr = theomax
                 endif
                 qdr = min(qdr,theomax)
                 lambda_intensity(inu) = xp * lambda_intensity(inu) + qdr
              endif
           else
              write(stdo,*) 'ERROR: Other grid than AMR/regular not yet implemented'
              stop
           endif
           !
        else
           !
           ! First order integration. Cell-based scheme.
           !
           ! Get the src and alp values
           !
           call lambda_get_src_alp(inu0,inu1,nrfreq,src,alp)
           !
           ! Now do the integration of the formal transfer equation
           !
           do inu=inu0,inu1
              !
              ! Compute the exp(-tau)
              !
              xp = alp(inu) * ray_ds
              if(xp.lt.0.d0) stop 7329
              if(xp.gt.1d-4) then
                 xp  = exp(-xp)
                 xp1 = 1.d0 - xp
              else
                 xp1 = xp
                 xp  = exp(-xp)
              endif
              !
              ! Now do the RT, including everything
              !
              lambda_intensity(inu) = xp * lambda_intensity(inu) + xp1 * src(inu)
           enddo
        endif
     elseif(camera_incl_stars.ne.0) then
        !
        ! Special situation: we are inside a star.
        ! So set the intensity to that of the star.
        !
        do inu=inu0,inu1
           lambda_intensity(inu) = find_starlight_interpol(camera_frequencies(inu), &
                                                    camera_istar)
        enddo
        !
     endif
####################
####################


!------------------------------------------------------------------------
!            INTEGRATE FORMAL TRANSFER EQUATION ALONG A RAY
!------------------------------------------------------------------------
subroutine lambda_integrate_ray(inu0,inu1,halt)
  implicit none
  integer :: inu0,inu1
  logical :: halt
  double precision :: s,ds
  doubleprecision :: snu_prev,anu_prev,jnu_prev,snu_curr,anu_curr,jnu_curr
  !
  ! Check
  !
  if(lambda_shift.gt.0.d0) then
     write(stdo,*) 'ERROR in Lambda operator: Shift>0.'
     stop
  endif
  if(.not.allocated(lambda_intensity)) then
     write(stdo,*) 'ERROR in Lambda operator: lambda_intensity not allocated.'
     stop
  endif
  !
  ! Shift position to outside the grid
  !
  ray_cart_x = ray_cart_x + lambda_shift * ray_dirx
  ray_cart_y = ray_cart_y + lambda_shift * ray_diry
  ray_cart_z = ray_cart_z + lambda_shift * ray_dirz
  s = lambda_shift
  !
  ! Start the intensity
  !
  lambda_intensity(inu0:inu1) = lambda_bgint(inu0:inu1)
  !
  ! Set defaults
  !
  snu_prev = 0.d0
  anu_prev = 0.d0
  jnu_prev = 0.d0
  snu_curr = 0.d0
  anu_curr = 0.d0
  jnu_curr = 0.d0
  !
  ! Check the dust scattering mode, because this only works under
  ! certain conditions.
  !
  if(scattering_mode.ge.2) then
     write(stdo,*) 'ERROR in Lambda module: The Lambda Operator only'
     write(stdo,*) '      works with dust scattering if the scattering'
     write(stdo,*) '      is isotropic.'
     stop
  endif
  if(scattering_mode.ge.1) then
     if(.not.allocated(mcscat_scatsrc_iquv)) stop 3005
  endif
  !
  ! If mirror symmetry is switched on, then we must assure that 
  ! z is above the midplane
  !
  if(amrray_mirror_equator) then
     if(ray_cart_z.eq.0.d0) then
        ray_cart_dirz = abs(ray_cart_dirz)
     elseif(ray_cart_z.lt.0.d0) then
        ray_cart_z    = -ray_cart_z
        ray_cart_dirz = -ray_cart_dirz
     endif
  endif
  !
  ! We start outside the domain, so all indices must be 0
  !
  ray_index     = 0
  ray_indexnext = 0
  !
  ! Now trace
  !
  arrived = .false.
  do while(.not.arrived)
     !
     ! Back up the current position
     !
     ray_prev_x     = ray_cart_x
     ray_prev_y     = ray_cart_y
     ray_prev_z     = ray_cart_z
     !
     ! Find new position
     !
     if(igrid_type.lt.100) then 
        !
        ! We use a regular or AMR grid
        !
        if(igrid_coord.lt.100) then
           !
           ! We use cartesian coordinates
           !
           call amrray_find_next_location_cart(ray_dsend,            &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived,              &
                levelnext=levelnext)
           !
           ! If we went through one of the boundaries of the
           ! domain, we might need to implement the thermal
           ! boundary, if present
           !
           if(incl_thermbc.ne.0) then
              if((ray_index.eq.0).and.(ray_indexnext.ne.0)) then
                 !
                 ! First determine which boundary we just went
                 ! through 
                 !
                 if(amrray_icross.ge.0) stop 3902  ! Self-consistency check
                 if(amrray_icross.eq.-1) then
                    bc_idir = 1
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-2) then
                    bc_idir = 1
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-3) then
                    bc_idir = 2
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-4) then
                    bc_idir = 2
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-5) then
                    bc_idir = 3
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-6) then
                    bc_idir = 3
                    bc_ilr  = 2
                 else
                    stop 4944
                 endif
                 !
                 ! Now set the intensity to the thermal
                 ! emission of the boundary, if the
                 ! thermal boundary is switched on
                 !
                 if(thermal_bc_active(bc_ilr,bc_idir)) then
                    do inu=1,lambda_nfreq
                       lambda_intensity(inu) = bplanck(thermal_bc_temp(bc_ilr,bc_idir), &
                                                lambda_frequencies(inu))
                    enddo
                 endif
              endif
           endif
           !
        elseif(igrid_coord.lt.200) then
           !
           ! We use spherical coordinates
           !
           call amrray_find_next_location_spher(ray_dsend,           &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived)
        else
           write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
           stop
        endif
        !
        ! Set the lambda_istar, which is used for treating finite-size
        ! stars (instead of point source stars)
        !
        lambda_istar = amrray_ispherehit
        !
     else
        write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
        stop
     endif
     !
     ! Path length
     !
     ds    = sqrt( (ray_cart_x-ray_prev_x)**2 + (ray_cart_y-ray_prev_y)**2 + (ray_cart_z-ray_prev_z)**2 )
     !
     ! Now do the integration along this path element
     !

     !
     ! Add information to the cell
     !
###############################################################
###############################################################
###############################################################
###############################################################
###############################################################
     !
     ! Increase s
     !
     s = s + ds
     !
     ! Now make next cell the current cell
     !
     ray_index = ray_indexnext
     !
     ! Do the same for the pointers.
     ! This is grid type dependent.
     !
     if(igrid_type.lt.100) then
        !
        ! AMR type grid
        !
        if(ray_index.gt.0) then
           !
           ! We are in a cell
           !
           if(amr_tree_present) then
              amrray_cell => amr_index_to_leaf(ray_index)%link
           else
              call amr_regular_get_ixyz(ray_index,amrray_ix_curr,amrray_iy_curr,amrray_iz_curr)
           endif
        else
           nullify(amrray_cell)
           amrray_ix_curr = -1
           amrray_iy_curr = -1
           amrray_iz_curr = -1
        endif
     else
        write(stdo,*) 'ERROR: This grid type not yet implemented'
        stop
     endif
     !
  enddo
  !
end subroutine lambda_integrate_ray

!------------------------------------------------------------------------
!                    RANDOM RAY LAMBDA OPERATOR
!------------------------------------------------------------------------
subroutine lambda_random_ray
  implicit none
end subroutine lambda_random_ray

end module lambda_module
